/************************************************************************
**
** This program is to test the generation of wave files
**
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "wave.h"

char data[] = "data";
char riff[] = "RIFF";
char fmt[] = "fmt ";
char wave[] = "WAVE";

WAVEFILE wavefile;

//-------------------------------------------------------------------------
// Special multiply routine to simulate a dsp multiply
//-------------------------------------------------------------------------

long dspMul(long x,long y)
{
	return ((x>>16) * (y >> 16)) << 1;
}

//------------------------------------------------------------------------------
// VCO class member functions
//------------------------------------------------------------------------------

vco::vco(int number_of_inputs)
{
	phaseaccum = 0x0l;	//initial value of phase accumulator
	inputs = new long *[number_of_inputs];
	maxinputs = number_of_inputs;
	memset(inputs,0,sizeof(long *) * number_of_inputs);	//zero memory
	ninputs=0;
}

vco::~vco()
{
	delete[] inputs;	//recover memory
}

int vco::Get(void)
{
	return (int)(phaseaccum >> 16);	//get upper 16 bits for output
}

void vco::Sync(void)
{
	phaseaccum = 0;
}

void vco::itterate(void)
{
	int i=0;

	output = phaseaccum;		//this could get more complicated later

	//--------------------------------------------------------------
	//next add all inputs into phase accumulator to get next value
	//--------------------------------------------------------------
	while(inputs[i] && (i < maxinputs)) phaseaccum += *inputs[i++];
}

long *vco::GetOutput()
{
	return &output;	//return pointer to vco output
}

int vco::SetInput(long *newinput)
{
	int retval = 0;	//inicate succuss prematurely

	if(ninputs != maxinputs)
		inputs[ninputs++] = newinput;
	else
		retval = -1;	//no more room for inputs
	return retval;
}

//-------------------------------------------------------------------------
// Member functions of multimode VCF
//-------------------------------------------------------------------------

mmfilter::mmfilter(int nf,int nq,int nv)
{
	z1=0;
	z2 = 0;
	lp = 0;
	bp = 0;
	hp = 0;
	freqinputs = new long *[nf];
	nfreqin = 0;
	maxfreqin = nf;
	memset(freqinputs,0,sizeof(long *) * nf);
	qinputs = new long *[nq];
	nqin = 0;
	maxqin = nq;
	memset(qinputs,0,sizeof(long *) * nq);
	vinputs = new long *[nv];
	nvin = 0;
	maxvin = nv;
	memset(vinputs,0,sizeof(long *) * nv);
}

mmfilter::~mmfilter()
{
	delete[] vinputs;		//free up allocated memory
	delete[] qinputs;
	delete[] freqinputs;
}

int mmfilter::SetVinput(long *v)
{
	int retval=0;

	if(nvin != maxvin)
		vinputs[nvin++] = v;
	else
		retval = -1;
	return retval;
}

int mmfilter::SetQinput(long *q)
{
	int retval = 0;

	if(nqin != maxqin)
		qinputs[nqin++] = q;
	else
		retval = -1;
	return retval;
}

int mmfilter::SetFinput(long *f)
{
	int retval = 0;

	if(nfreqin != maxfreqin)
		freqinputs[nfreqin++] = f;
	else
		retval = -1;
	return retval;
}

long *mmfilter::GetOutput(int s)	//select an output
{
	long *retval;

	switch(s)
	{
		case MMFILTER_LOWPASS:
			retval = &lp;
			break;
		case MMFILTER_HIGHPASS:
			retval = &hp;
			break;
		case MMFILTER_BANDPASS:
			retval = &bp;
			break;
		default:
			retval = 0;
			break;
	}
	return retval;
}

void mmfilter::itterate(void)
{
	//--------------------------------
	// accumulate inputs
	//--------------------------------
	long Q=0;
	long Vin=0;
	long F=0;
	int i;

	for(i=0;freqinputs[i] && (i < maxfreqin);++i) F += *freqinputs[i];
	for(i=0;qinputs[i] && (i < maxqin);++i) Q += *qinputs[i];
	for(i=0;vinputs[i] && (i < maxvin);++i) Vin += *vinputs[i];
	//--------------------------------
	// oh joy, do filter calculation
	//--------------------------------
	lp = z2 + dspMul(F,z1);
	hp = - (lp + dspMul(Q,z1 - Vin));
	bp = dspMul(F,hp) + z1;
	z1 = bp;
	z2 = lp;
}

//------------------------------------------------------------------------
// Member functions for VCA module
//------------------------------------------------------------------------

vca::vca(int nv,int nc)
{
	vinputs = new long *[nv];	//allocate memory for signal input pointers
	nvin = 0;
	maxvin = nv;
	cinputs = new long *[nc];	//allocate memory for control input pointers
	ncin = 0;
	maxcin = nc;
	memset(vinputs,0,sizeof(long *) * nv);
	memset(cinputs,0,sizeof(long *) * nc);
}

vca::~vca()
{
	delete[] cinputs;
	delete[] vinputs;
}

int vca::SetVinput(long *v)
{
	int retval = 0;
	if(nvin != maxvin)
		vinputs[nvin++] = v;
	else
		retval = -1;
	return retval;
}

int vca::SetCinput(long *c)
{
	int retval = 0;
	if(ncin != maxcin)
		cinputs[ncin++] = c;
	else
		retval = -1;
	return retval;
}

long *vca::GetOutput(void)
{
	return &output;		//return pointer to output value
}

void vca::itterate(void)
{
	long input=0;
	long control=0;
	int i=0;
	while(vinputs[i]) input += *vinputs[i++];
	i=0;
	while(cinputs[i]) control += *cinputs[i++];
	output = dspMul(input,control);	//just multiply
}

//------------------------------------------------------------------------
// Attack-Decay-Sustain-Release Envelope Generator
//------------------------------------------------------------------------

adsr::adsr(int na,int nd,int ns,int nr)
{
	output = 0;
	z = 0;
	attack = new long *[na];
	memset(attack,0,sizeof(long *)*na);
	nattack = 0;
	maxattack = na;
	decay = new long *[nd];
	memset(decay,0,sizeof(long *)*nd);
	ndecay = 0;
	maxdecay = nd;
	sustain = new long *[ns];
	memset(sustain,0,sizeof(long *)*ns);
	nsustain = 0;
	maxsustain = ns;
	release = new long *[nr];
	memset(release,0,sizeof(long *)*nr);
	nrelease = 0;
	maxrelease = nr;
	*gate = 0;
	*trig = 0;
   state = 0;
}

adsr::~adsr()
{
	delete[] release;
	delete[] sustain;
	delete[] decay;
	delete[] attack;
}

int adsr::SetAttack(long *v)
{
	int retval = 0;
	if(nattack != maxattack)
		attack[nattack++] = v;
	else
		retval = -1;
	return retval;
}

int adsr::SetDecay(long *v)
{
	int retval = 0;
	if(ndecay != maxdecay)
		decay[ndecay++] = v;
	else
		retval = -1;
	return retval;
}

int adsr::SetSustain(long *v)
{
	int retval = 0;
	if(nsustain != maxsustain)
		sustain[nsustain++] = v;
	else
		retval = -1;
	return retval;
}

int adsr::SetRelease(long *v)
{
	int retval = 0;
	if(nrelease != maxrelease)
		release[nrelease++] = v;
	else
		retval = -1;
	return retval;
}

void adsr::SetGate(int *g)
{
	gate = g;
}

void adsr::SetTrig(int *t)
{
	trig = t;
}

long *adsr::GetOutput(void)
{
	return &output;
}

void adsr::itterate(void)
{
	long A=0;
	long D=0;
	long S=0;
	long R=0;
	long I;
	long W;
	int i;

	for(i=0;attack[i] && (i < maxattack);++i) A += *attack[i];
	for(i=0;decay[i] && (i < maxdecay);++i) D += *decay[i];
	for(i=0;sustain[i] && (i < maxsustain);++i) S += *sustain[i];
	for(i=0;release[i] && (i < maxrelease);++i) R += *release[i];
	if(gate && trig)	//are these pointers valid?
	{
		//do adsr thing
		switch(state)
		{
			case 0:			//quiesent state
				if(*trig && *gate)
				{
					state = 1;
					W = A;				//go at attack rate
					I = 0x7ffffffl;	//attack up to largest value
				}
				else if(!*trig && gate)
				{
					state = 2;
					W = D;		//go at decay rate
					I = S; 		//go to sustain level
				}
				else
				{
					W = R;
					I = 0l;
				}
				break;
			case 1:
				if(*gate) //while gate still active
				{
					W = A;
					I = 0x7fffffffl;
				}
				else
				{
					state = 0;
					W = R;
					I = 0;
				}
				break;
			case 2:
				if(*gate)	//while gate still active
				{
					if(*trig)
					{
						state = 1;
						W = A;
						I = 0x7fffffffl;
					}
					else
					{
						W = D;
						I = S;
					}
				}
				break;
		}	//end of switch state
		//-------------------------------------------------------------
		// do one pole digital filter
		//-------------------------------------------------------------
		output = I-z;
		if((state == 1) && (labs(output) < 2000l))
			state = 2;	//end of attack sequence
		output = dspMul(W,output) + z;
      z = output;
	}	//end of if(gate && trig)
}

//------------------------------------------------------------------------
// Output Module
//------------------------------------------------------------------------

Output::Output(int ni,FILE *out)
{
	outfile = out;
	inputs = new long *[ni];
   memset(inputs,0,sizeof(long *) * ni);
	ninputs = 0;
	maxinputs = ni;
}

Output::~Output()
{
	delete[] inputs;
}

int Output::SetInput(long *in)
{
	int retval=0;
	if(ninputs != maxinputs)
		inputs[ninputs++] = in;
	else
		retval = -1;
	return retval;
}

void Output::itterate(void)
{
	long out=0;
	int i=0;
	while(inputs[i] && i < maxinputs) out += *inputs[i++];
	i = int(out >> 24);
	i ^= 0x80;		//wave file '0'
   fputc(i,outfile);
}

//------------------------------------------------------------------------
// Dac Modules
//------------------------------------------------------------------------

Dac::Dac()
{
	output = 0;
	gate = 0;
	trig = 0;
	oldtrig = 0;
}

Dac::~Dac()
{
}

void Dac::SetFreq(double F)
{
	output = long((F / 22050.0) * 4294967296.0);
}

void Dac::SetTrig(void)
{
	trig = 1;
}

void Dac::SetGate(int flag)
{
	gate = flag?1:0;
}

long *Dac::GetOutput(void)
{
	return &output;
}

int *Dac::GetGate(void)
{
	return &gate;
}

int *Dac::GetTrig(void)
{
	return &trig;
}

void Dac::itterate(void)
{
	if(oldtrig && trig)
	{
		trig = 0;
		oldtrig = 0;
	}else	if(trig && !oldtrig) oldtrig = 1;
}

//------------------------------------------------------------------------
// program to test modules
//------------------------------------------------------------------------

main()
{
	//--------------------------------------------------------------------
	// Set up wave file
	//--------------------------------------------------------------------
	FILE *out;
	char *s = new char[256];
	printf("Enter filename\n");
	gets(s);
	if((out = fopen(s,"wb")) == NULL)
	{
		printf("Could not open %s for output\n",s);
		exit(0);
	}
	printf("Initialize wave table header\n");
	memcpy(wavefile.riff.rID,riff,4);
	wavefile.riff.rLen = 22086l;
	memcpy(wavefile.wave.wID,wave,4);
	memcpy(wavefile.wave.fID,fmt,4);
	wavefile.wave.fLen = 16;
	wavefile.wave.wFormatTag=1;
	wavefile.wave.nChannels = 1;
	wavefile.wave.nSamplesPerSec = 22050l;
	wavefile.wave.nAvgBytesPerSec = 22050l;
	wavefile.wave.nBlockAlign = 1;
	wavefile.wave.FormatSpecific = 8;
	memcpy(wavefile.data.dID,data,4);
	wavefile.data.dLen = 22050;	//one second of sound
	int i;
	char *p = (char *)&wavefile;
	int x = sizeof(WAVEFILE);
	for(i=0;i<x;++i,++p)
		fputc(*p,out);	//output header

	//------------------------------------------------------------------
	// allocate synthsizer modules
	//------------------------------------------------------------------
	vco *VCO = new vco(4);
	Dac *DAC = new Dac();
	Output *VO = new Output(4,out);
	pot *P1 = new pot();
	pot *P2 = new pot();
	pot *P3 = new pot();
	vca *VCA = new vca(4,4);
	mmfilter *VCF = new mmfilter(4,4,4);

	//-------------------------------------------------------------------
	// patche them together
	//-------------------------------------------------------------------
	VCO->SetInput(DAC->GetOutput());	//connect output of DAC to VCO
	VCA->SetVinput(VCO->GetOutput());	//connect VCO to VCA
	VCA->SetCinput(P1->GetOutput());		//connect POT to VCA control
	VCF->SetVinput(VCA->GetOutput());	//connect filter to VCA
	VCF->SetQinput(P2->GetOutput());		//connect pot to Q input
	VCF->SetFinput(P3->GetOutput());
	VO->SetInput(VCF->GetOutput(MMFILTER_LOWPASS));	//connect VCF to output module

	//************************************************************
	// perform interations on modules
	//************************************************************
	DAC->SetFreq(440.0);
	P1->SetPot(0x20000000l);
	P2->SetPot(0x20000000l);		//moderately low q
	P3->SetPot(0x60000000l);		//moderate frequency
	for(i=0;i<22050;++i)
	{
		DAC->itterate();
		VCO->itterate();
		VCA->itterate();
		VCF->itterate();
		VO->itterate();
	}
	fclose(out);
	printf("DONE\n");
	delete[] s;
}